#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `graphtransliterator` package."""
import graphtransliterator
import pytest
import re
import yaml
from graphtransliterator import GraphTransliterator
from graphtransliterator.core import _easyreading_rule
from graphtransliterator import process, validate
from graphtransliterator.rules import OnMatchRule, TransliterationRule, WhitespaceRules
from graphtransliterator.graphs import DirectedGraph

from graphtransliterator.exceptions import (
    NoMatchingTransliterationRuleException,
    UnrecognizableInputTokenException,
    AmbiguousTransliterationRulesException,
)

yaml_for_test = r"""
tokens:
  a: [token, class1]
  b: [token, class2]
  u: [token]
  ' ': [wb]
rules:
  a: A
  b: B
  <wb> u: \N{DEVANAGARI LETTER U}
onmatch_rules:
  -
    <class1> + <class2>: ","
  -
    <class1> + <token>: \N{DEVANAGARI SIGN VIRAMA}
whitespace:
  default: ' '
  token_class: 'wb'
  consolidate: true
metadata:
  author: Author Name
"""


def test_GraphTransliterator_from_YAML():
    """Test YAML loading of GraphTransliterator."""
    good_yaml = """
      tokens:
        a: [class1]
        ' ': [wb]
      rules:
        a: A
      whitespace:
        default: ' '
        consolidate: true
        token_class: wb
    """
    assert GraphTransliterator.from_yaml(good_yaml)
    bad_yaml = """
      tokens:
        a: class1
        ' ': wb
      rules:
        a: A
      whitespace:
        default: ' '
        consolidate: true
        token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)

    bad_yaml = """
          rules:
            a: A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)
    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)
    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            a: A
    """
    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            b: A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)

    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            (b) a: A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)

    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            a (b): A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)

    bad_yaml = """
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            a <class_nonexisting>: A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)

    # test for bad tokens
    bad_yaml = """
          tokens: '7'
          rules:
            a <class_nonexisting>: A
          whitespace:
            default: ' '
            consolidate: true
            token_class: wb
    """
    with pytest.raises(ValueError):
        GraphTransliterator.from_yaml(bad_yaml)


def test_graphtransliterator_process():
    """Test graphtransliterator proccessing of rules."""

    data = yaml.safe_load(yaml_for_test)

    assert process._process_rules({"a": "A"})[0]["tokens"] == ["a"]
    assert process._process_rules({"a": "A"})[0]["production"] == "A"
    assert (
        process._process_onmatch_rules(data["onmatch_rules"])[0]["prev_classes"][0]
        == "class1"
    )
    assert (
        process._process_onmatch_rules(data["onmatch_rules"])[0]["next_classes"][0]
        == "class2"
    )


def test_graphtransliterator_models():
    """Test internal models."""
    tr = TransliterationRule(
        production="A",
        prev_classes=None,
        prev_tokens=None,
        tokens=["a"],
        next_tokens=None,
        next_classes=None,
        cost=1,
    )
    assert tr.cost == 1
    # assert TransliteratorOutput([tr], 'A').output == 'A'
    assert OnMatchRule(prev_classes=["class1"], next_classes=["class2"], production=",")
    assert WhitespaceRules(default=" ", token_class="wb", consolidate=False)


def test_graphtransliterator_structures():
    graph = DirectedGraph()

    assert len(graph.node) == 0
    assert len(graph.edge) == 0

    graph.add_node({"type": "test1"})
    graph.add_node({"type": "test2"})
    assert graph.node[0]["type"] == "test1"
    assert graph.node[1]["type"] == "test2"

    graph.add_edge(0, 1, {"type": "edge_test1"})
    assert graph.edge[0][1]["type"] == "edge_test1"
    assert type(graph.to_dict()) == dict

    # edge tail not in graph
    with pytest.raises(ValueError):
        graph.add_edge(0, 7, {})
    # edge head not in graph
    with pytest.raises(ValueError):
        graph.add_edge(7, 0, {})

    # invalid edge data
    with pytest.raises(ValueError):
        graph.add_edge(0, 1, "not a dict")

    # invalid edge head type
    with pytest.raises(ValueError):
        graph.add_edge("zero", 1)
    # invalid edge tail type
    with pytest.raises(ValueError):
        graph.add_edge(1, "zero")
    # invalid node data
    with pytest.raises(ValueError):
        graph.add_node("Not a dict")


def test_graphtransliterator_validate_settings():
    """Test graph transliterator validation of settings."""
    settings = yaml.safe_load(yaml_for_test)
    # check for bad tokens
    settings["tokens"] = "bad token"
    with pytest.raises(ValueError):
        validate.validate_settings(
            settings["tokens"],
            settings["rules"],
            settings["onmatch_rules"],
            settings["whitespace"],
            {},
        )


def test_GraphTransliterator_transliterate(tmpdir):
    """Test GraphTransliterator transliterate."""
    YAML = r"""
    tokens:
        a: [class_a]
        b: [class_b]
        c: [class_c]
        " ": [wb]
        d: []
        Aa: [contrained_rule]
    rules:
        a: A
        b: B
        <class_c> <class_c> a: A(AFTER_CLASS_C_AND_CLASS_C)
        (<class_c> b) a: A(AFTER_B_AND_CLASS_C)
        (<class_c> b b) a a: AA(AFTER_BB_AND_CLASS_C)
        a <class_c>: A(BEFORE_CLASS_C)
        a b (c <class_b>): AB(BEFORE_C_AND_CLASS_B)
        c: C
        c c: C*2
        a (b b b): A(BEFORE_B_B_B)
        d (c <class_a>): D(BEFORE_C_AND_CLASS_A)
        (b b) a: A(AFTER_B_B)
        <wb> Aa: A(ONLY_A_CONSTRAINED_RULE)
    onmatch_rules:
        -
            <class_a> <class_b> + <class_a> <class_b>: "!"
        -
            <class_a> + <class_b>: ","
    whitespace:
        default: ' '
        consolidate: True
        token_class: wb
    """
    gt = GraphTransliterator.from_yaml(YAML)
    # rules with single token
    assert gt.transliterate("a") == "A"
    # rules with multiple tokens
    assert gt.transliterate("aa") == "AA"
    # rules with multiple tokens (for rule_key)
    assert gt.transliterate("cc") == "C*2"
    # # rules with multiple tokens overlapping end of tokens
    # assert gt.transliterate('c') == 'C'

    # rules with prev class
    assert gt.transliterate("ca") == "CA"
    # rules with prev class and prev token
    assert gt.transliterate("dca") == "D(BEFORE_C_AND_CLASS_A)CA"
    # rules with prev class and prev tokens
    assert gt.transliterate("cbba") == "CBBA(AFTER_B_B)"
    # rules with next class
    assert gt.transliterate("ac") == "A(BEFORE_CLASS_C)C"
    # rules with next class and next tokens
    assert gt.transliterate("acb") == "A(BEFORE_CLASS_C)CB"
    # rules with onmatch rule of length 1
    assert gt.transliterate("ab") == "A,B"
    # rules that only have constraints on first element
    assert gt.transliterate("Aa") == "A(ONLY_A_CONSTRAINED_RULE)"
    # test whitespace consolidation
    assert gt.transliterate(" a") == "A"
    # test whitespace consolidation following
    assert gt.transliterate("a ") == "A"

    # rules with longer onmatch rules
    assert gt.transliterate("abab") == "A,B!A,B"

    # test last_matched_input_tokens
    assert gt.last_input_tokens == [" ", "a", "b", "a", "b", " "]
    # test last_matched_tokens
    assert gt.last_matched_rule_tokens == [["a"], ["b"], ["a"], ["b"]]

    # test last_matched_rules
    assert len(gt.last_matched_rules) == 4

    # test serialization
    assert gt.serialize()["graph"]["edge"]
    assert re.match(r"\d+\.\d+\.\d+$", gt.serialize()["graphtransliterator_version"])
    assert (
        gt.serialize()["graphtransliterator_version"] == graphtransliterator.__version__
    )


def test_version():
    """Tests to make sure version is not a mess due to Black formatting"""

    assert re.match(r"\d+\.\d+\.\d+$", graphtransliterator.__version__)


def test_match_all():
    """Test GraphTransliterator transliterate."""
    YAML = r"""
    tokens:
        a: [class_a]
        " ": [wb]
    rules:
        a: A
        a a: A*2
    whitespace:
        default: ' '
        consolidate: True
        token_class: wb
    """
    gt = GraphTransliterator.from_yaml(YAML)
    assert gt.rules[0].cost < gt.rules[1].cost

    tokens = gt.tokenize("aa")
    assert gt.match_at(1, tokens, match_all=False) == 0
    assert gt.match_at(1, tokens, match_all=True) == [0, 1]


def test_GraphTransliterator(tmpdir):
    """Test GraphTransliterator."""
    yaml_str = r"""
    tokens:
      a: [token, class1]
      b: [token, class2]
      u: [token]
      ' ': [wb]
    rules:
      a: A
      b: B
      <wb> u: \N{DEVANAGARI LETTER U}
    onmatch_rules:
      -
        <class1> + <class2>: ","
      -
        <class1> + <token>: \N{DEVANAGARI SIGN VIRAMA}
    whitespace:
      default: ' '
      token_class: 'wb'
      consolidate: true
    metadata:
      author: Author
    """

    input_dict = yaml.safe_load(yaml_str)
    assert "a" in GraphTransliterator.from_dict(input_dict).tokens.keys()
    gt = GraphTransliterator.from_dict(input_dict)
    assert gt.onmatch_rules[0].production == ","
    assert gt.tokens
    assert gt.rules
    assert gt.whitespace
    assert gt.whitespace.default
    assert gt.whitespace.token_class
    assert gt.whitespace.consolidate
    assert gt.metadata["author"] == "Author"
    assert type(gt.graph) == DirectedGraph
    yaml_file = tmpdir.join("yaml_test.yaml")
    yaml_filename = str(yaml_file)
    yaml_file.write(yaml_str)

    assert yaml_file.read() == yaml_str

    assert GraphTransliterator.from_yaml_file(yaml_filename)

    assert len(set(GraphTransliterator.from_dict(input_dict).tokens)) == 4

    assert GraphTransliterator.from_yaml(yaml_str).transliterate("ab") == "A,B"
    assert (
        GraphTransliterator.from_yaml_file(yaml_filename).transliterate("ab") == "A,B"
    )
    assert (
        GraphTransliterator.from_dict(
            {
                "tokens": {"a": ["class_a"], "b": ["class_b"], " ": ["wb"]},
                "onmatch_rules": [{"<class_a> + <class_b>": ","}],
                "whitespace": {
                    "default": " ",
                    "token_class": "wb",
                    "consolidate": True,
                },
                "rules": {"a": "A", "b": "B"},
            },
            raw=True,
        ).transliterate("ab")
        == "A,B"
    )


def test_GraphTransliterator_ignore_errors():
    # if ignore_errors is not set and no matching transliteration rule
    # raise NoMatchingTransliterationRule exception
    yaml_str = """
        tokens:
           a: [class1]
           b: [class1]
           ' ': [wb]
        rules:
           a a: B2
           b: B
        whitespace:
           default: ' '
           consolidate: true
           token_class: wb
           """
    # check that ignore_errors works
    assert (
        GraphTransliterator.from_yaml(yaml_str, ignore_errors=True).transliterate("a")
        == ""
    )

    with pytest.raises(NoMatchingTransliterationRuleException):
        gt = GraphTransliterator.from_yaml(yaml_str, ignore_errors=False)
        assert gt.ignore_errors is False
        gt.transliterate("a")

    with pytest.raises(UnrecognizableInputTokenException):
        gt = GraphTransliterator.from_yaml(yaml_str, ignore_errors=False)
        assert gt.ignore_errors is False
        gt.transliterate("!")

    gt = GraphTransliterator.from_yaml(yaml_str, ignore_errors=True)
    assert gt.ignore_errors is True
    assert gt.tokenize("b!b") == [" ", "b", "b", " "]
    assert gt.transliterate("b!b") == "BB"

    with pytest.raises(UnrecognizableInputTokenException):
        gt = GraphTransliterator.from_yaml(yaml_str, ignore_errors=False)
        assert gt.ignore_errors is False
        gt.transliterate("b!")

    # # test ignore_errors keyword value checking on init
    # with pytest.raises(ValueError):
    #     GraphTransliterator.from_yaml(yaml_str, ignore_errors="maybe")
    # test ignore_errors keyword property

    # test ignore_errors setter and property
    gt.ignore_errors = True
    assert gt.ignore_errors is True
    gt.ignore_errors = False
    assert gt.ignore_errors is False
    # test ignore_errors setter exception handling
    # with pytest.raises(ValueError):
    #     gt.ignore_errors = "Maybe"


def test_GraphParser_check_ambiguity():
    """ Test for rules that can both match the same thing."""

    yaml_for_test = r"""
        tokens:
          a: [token, class1, class2]
          b: [token, class1, class2]
          ' ': [wb]
        rules:
          a <class1>: A<class1> # these should be ambiguous
          a <class2>: A<class2>

          <class1> a: <class1>A  # these should be ambiguous
          <class2> a: <class2>A # these should be ambiguous

          (<class1> b) a (b <class2>): A # ambigous
          (<class2> b) a (b <class1>): A # ambiguous
          a: A # not ambiguous
        whitespace:
          default: ' '
          token_class: 'wb'
          consolidate: true
        """
    with pytest.raises(AmbiguousTransliterationRulesException):
        GraphTransliterator.from_yaml(yaml_for_test, check_for_ambiguity=True)
    # check that ambiguity matches if rules are of different shape
    yaml = """
        tokens:
          a: []
          ' ': [wb]
        rules:
          <wb> a: _A
          a <wb>: A_
          a: a
          ' ': ' '
        whitespace:
          default: " "        # default whitespace token
          consolidate: true  # whitespace should be consolidated
          token_class: wb     # whitespace token class
        """
    with pytest.raises(AmbiguousTransliterationRulesException):
        GraphTransliterator.from_yaml(yaml, check_for_ambiguity=True)


def test_GraphTransliterator_types():
    """Test internal types."""
    pr = TransliterationRule(
        production="A",
        prev_classes=None,
        prev_tokens=None,
        tokens=["a"],
        next_tokens=None,
        next_classes=None,
        cost=1,
    )
    assert pr.cost == 1
    #    assert TransliteratorOutput([pr], 'A').output == 'A'
    assert OnMatchRule(prev_classes=["class1"], next_classes=["class2"], production=",")
    assert WhitespaceRules(default=" ", token_class="wb", consolidate=False)

    graph = DirectedGraph()

    assert len(graph.node) == 0
    assert len(graph.edge) == 0

    graph.add_node({"type": "test1"})
    graph.add_node({"type": "test2"})
    assert graph.node[0]["type"] == "test1"
    assert graph.node[1]["type"] == "test2"

    graph.add_edge(0, 1, {"type": "edge_test1"})
    assert graph.edge[0][1]["type"] == "edge_test1"


def test_GraphTransliterator_productions():
    """Test productions."""
    tokens = {"ab": ["class_ab"], " ": ["wb"]}
    whitespace = {"default": " ", "token_class": "wb", "consolidate": True}
    rules = {"ab": "AB", " ": "_"}
    settings = {"tokens": tokens, "rules": rules, "whitespace": whitespace}
    assert set(GraphTransliterator.from_dict(settings).productions) == set(["AB", "_"])


def test_GraphTransliterator_pruned_of():
    gt = GraphTransliterator.from_yaml(
        """
            tokens:
               a: [class1]
               b: [class2]
               ' ': [wb]
            rules:
               a: A
               b: B
            whitespace:
               default: ' '
               consolidate: true
               token_class: wb
        """
    )
    assert len(gt.rules) == 2
    assert len(gt.pruned_of("B").rules) == 1
    assert gt.pruned_of("B").rules[0].production == "A"
    assert gt.pruned_of(["A", "B"])  # if no rules present will still work


def test_GraphTransliterator_easy_reading():
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], [], ["a"], [], ["class_a"], 0)
        )
        == "<class_a> a <class_a>"
    )
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], [], ["a"], [], ["class_a"], 0)
        )
        == "<class_a> a <class_a>"
    )
    assert (
        _easyreading_rule(TransliterationRule("", [], ["b"], ["a"], ["b"], [], 0))
        == "(b) a (b)"
    )
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], ["b"], ["a"], ["b"], ["class_a"], 0)
        )
        == "(<class_a> b) a (b <class_a>)"
    )


def test_GraphTransliterator_graph():
    """Test graph."""
    tokens = {"ab": ["class_ab"], " ": ["wb"]}
    whitespace = {"default": " ", "token_class": "wb", "consolidate": True}
    rules = {"ab": "AB", " ": "_"}
    settings = {"tokens": tokens, "rules": rules, "whitespace": whitespace}
    gt = GraphTransliterator.from_dict(settings)
    assert gt._graph
    assert gt._graph.node[0]["type"] == "Start"  # test for Start
    assert gt



# -*- coding: utf-8 -*-

"""
Transliteration exceptions used in Graph Transliterator.
"""


class GraphTransliteratorException(Exception):
    """
    Base exception class. All Graph Transliterator-specific exceptions should
    subclass this class.
    """


class AmbiguousTransliterationRulesException(GraphTransliteratorException):
    """
    Raised when multiple transliteration rules can match the same pattern.
    Details of ambiguities are given in a :func:`logging.warning`.
    """


class NoMatchingTransliterationRuleException(GraphTransliteratorException):
    """
    Raised when no transliteration rule can be matched at a particular
    location in the input string's tokens. Details of the location are given
    in a :func:`logging.warning`.
    """


class UnrecognizableInputTokenException(GraphTransliteratorException):
    """
    Raised when a character in the input string does not correspond to any
    tokens in the GraphTransliterator's token settings. Details of the location
    are given in a :func:`logging.warning`.
    """



# -*- coding: utf-8 -*-

"""
graphtransliterator
~~~~~~~~~~~~~~~~~~
Graph Transliterator is a Python module enabling rule-based transliteration
of the tokens of an input string to an output string. It can be used to
transliterate the symbols of one language into those of another, or to
convert between transliteration systems.

:copyright: © 2019 Michigan State University.
:license: MIT, see LICENSE.rst for more details.
"""

__author__ = """A. Sean Pue"""
__email__ = "pue@msu.edu"
__version__ = "0.2.14"

# Core classes
from .core import GraphTransliterator  # noqa

# Graphs
from .graphs import DirectedGraph  # noqa

# Rules
from .rules import TransliterationRule, OnMatchRule, WhitespaceRules  # noqa

# Exceptions
from .exceptions import (
    GraphTransliteratorException,
    AmbiguousTransliterationRulesException,
    NoMatchingTransliterationRuleException,
    UnrecognizableInputTokenException,
)  # noqa

from .validate import validate_easyreading_settings, validate_settings

__all__ = [
    # core
    "GraphTransliterator",
    # exceptions
    "AmbiguousTransliterationRulesException",
    "GraphTransliteratorException",
    "NoMatchingTransliterationRuleException",
    "UnrecognizableInputTokenException",
    # graphs
    "DirectedGraph",
    # rules
    "TransliterationRule",
    "OnMatchRule",
    "WhitespaceRules",
    # validate
    "validate_easyreading_settings",
    "validate_settings",
]



# -*- coding: utf-8 -*-
"""Functions to process "easy reading" settings of GraphTransliterator."""

import re


# ----------- process settings (main method) ----------


def _process_easyreading_settings(settings):
    """
    Extract information from "easy reading" settings.

    Returns
    -------
    `dict` keyed by:
      `tokens`: dict of {str: list of str}
      `rules`:  list of dict {'prev_classes', 'prev_tokens', 'tokens',
                'next_tokens', 'next_classes'}
      `whitespace': dict {'default', 'token_class', 'consolidate'}
      `onmatch_rules`: list of dict {
                       'prev_classes', 'next_classes', 'production'}
    """

    processed_settings = {
        "tokens": settings.get("tokens"),
        "rules": _process_rules(settings.get("rules")),
        "onmatch_rules": _process_onmatch_rules(settings.get("onmatch_rules")),
        "whitespace": settings.get("whitespace"),
        "metadata": settings.get("metadata", {}),
    }
    return processed_settings


# ----------- process rules ----------

# added \n and \r explicitly as validator does not accept re.DOTALL

RULE_RE = re.compile(
    r"(?:\(((?:\s?<.+?>\s+)+)?(.+?)\s?\) |((?:\s?<.+?>\s+)+)?)?"
    r"((?:\n|\r|.)+?)"
    r"(?:"
    r"\s+(?:\((.+?)((?:\s+<.+?>?)?)\)$)"
    r"|"
    r"((?:\s+<.+?>)+)"
    r"|"
    r"$)"
)


def _process_rule(key, value):
    """Convert key and value of a "easy reading" rule into a dict."""

    rule = {"production": value}  # needs to be dict for Cerberus

    match = RULE_RE.match(key)
    assert match, "Rule pattern is not valid: %s" % key

    if match.group(1):
        rule["prev_classes"] = re.findall("<(.+?)>", match.group(1))
    if match.group(2):
        rule["prev_tokens"] = match.group(2).split(" ")
    if match.group(3):
        rule["prev_classes"] = re.findall("<(.+?)>", match.group(3))
    if match.group(4) == " ":
        rule["tokens"] = [" "]
    else:
        rule["tokens"] = [x for x in match.group(4).split(" ") if x != ""]
    if match.group(5):
        rule["next_tokens"] = match.group(5).split(" ")
    if match.group(6):
        rule["next_classes"] = re.findall("<(.+?)>", match.group(6))
    if match.group(7):
        rule["next_classes"] = re.findall("<(.+?)>", match.group(7))

    return rule


def _process_rules(easyreading_rules):
    """Processes "easy reading" rules dict into a list of dict."""

    return [_process_rule(key, value) for key, value in easyreading_rules.items()]


# ----------- process onmatch_rules -----------


ONMATCH_RE = re.compile(
    r"^(" r"(?:<[^+< \s]+>\s*)+" r")" r"\+" r"(" r"(?:\s*<[^+<]+>)+)\s*$"
)

ONMATCH_CLASS_RE = re.compile(r"(?<=<)[^+< \s]+(?=>)")


def _process_onmatch_rules(easyreading_onmatch_rules):
    """Process "easyreading" onmatch_rules (list of dict) into list of dict."""

    if not easyreading_onmatch_rules:
        return []

    processed_onmatch_rules = []

    for item in easyreading_onmatch_rules:

        assert len(item) == 1, "onmatch_rule has too many values: %s" % item

        (rule_easyreading, _production) = list(item.items())[0]

        match = ONMATCH_RE.match(rule_easyreading)

        prev_str = match.group(1)
        next_str = match.group(2)

        _prev_classes = ONMATCH_CLASS_RE.findall(prev_str)
        _next_classes = ONMATCH_CLASS_RE.findall(next_str)

        onmatch_rule_processed = {
            "prev_classes": _prev_classes,
            "next_classes": _next_classes,
            "production": _production,
        }

        processed_onmatch_rules.append(onmatch_rule_processed)

    return processed_onmatch_rules



# -*- coding: utf-8 -*-

"""
Graph Transliterator rule classes.
"""

from collections import namedtuple


class TransliterationRule(
    namedtuple(
        "TransliterationRule",
        [
            "production",
            "prev_classes",
            "prev_tokens",
            "tokens",
            "next_tokens",
            "next_classes",
            "cost",
        ],
    )
):
    """
    A transliteration rule containing the specific match conditions and
    string output to be produced, as well as the rule's cost.

    Attributes
    ----------
    production: `str`
        Output produced on match of rule
    prev_classes: `list` of `str`, or `None`
        List of previous token classes to be matched before tokens or,
        if they exist, `prev_tokens`
    prev_tokens: `list` of `str`, or `None`
        List of tokens to be matched before `tokens`
    tokens: `list` of `str`
        List of tokens to match
    next_tokens: `list` of `str`, or `None`
        List of tokens to match after `tokens`
    next_classes: `list` of `str`, or `None`
        List of tokens to match after `tokens` or, if they exist, `next_tokens`
    cost: `float`
        Cost of the rule, where less specific rules are more costly
    """

    __slots__ = ()


class OnMatchRule(
    namedtuple("OnMatchRule", ["prev_classes", "next_classes", "production"])
):
    """
    Rules about adding text between certain combinations of matched rules.

    When a translation rule has been found and before its production is added
    to the output, the productions string of an OnMatch rule is added if
    previously matched tokens and current tokens are of the specified classes.

    Attributes
    ----------
    prev_classes: `list` of `str`
        List of previously matched token classes required
    next_classes: `list` of `str`
        List of current and following token classes required
    production: `str`
        String to added before current rule
    """

    __slots__ = ()


class WhitespaceRules(
    namedtuple("Whitespace", ["default", "token_class", "consolidate"])
):
    """
    Whitespace rules of GraphTransliterator.

    Attributes
    ----------
    default: `str`
        Default whitespace token
    token_class: `str`
        Whitespace token class
    consolidate: `bool`
        Consolidate consecutive whitespace tokens and render as a single
        instance of the specified default whitespace token.
    """

    __slots__ = ()



# -*- coding: utf-8 -*-

"""
Graph data structure used in Graph Transliterator.
"""


class DirectedGraph:
    """
    A very basic dictionary- and list-based directed graph. Nodes are a list of
    dictionaries of node data. Edges are nested dictionaries keyed from the
    head -> tail -> edge properties. An edge list is maintained. Can be
    exported as a dictionary.

    Attributes
    ----------
    node : `list` of `dict`
        List of node attributes
    edge : `dict` of {`int`: `dict` of {`int`: `dict`}}
        Mapping from head to tail of edge, holding edge data
    edge_list : `list` of `tuple` of (`int`, `int`)
        List of head and tail of each edge

    Examples
    -------
    >>> from graphtransliterator import DirectedGraph
    >>> DirectedGraph()
    <graphtransliterator.graphs.DirectedGraph object at 0x106d1c908>
    """

    __slots__ = "edge", "node", "edge_list"

    def __init__(self):
        self.edge = {}
        self.node = []
        self.edge_list = []

    def add_edge(self, head, tail, edge_data={}):
        """
        Add an edge to a graph and return its attributes as dict.

        Parameters
        ----------
        head: `int`
            Index of head of edge
        tail: `int`
            Index of tail of edge
        edge_data: `dict`, default {}
            Edge data

        Returns
        -------
        dict
            Data of created edge

        Raises
        ------
        ValueError
            Invalid ``head`` or ``tail``, or ``edge_data`` is not a dict.

        Examples
        --------
        >>> g = DirectedGraph()
        >>> g.add_node()
        (0, {})
        >>> g.add_node()
        (1, {})
        >>> g.add_edge(0,1, {'data_key_1': 'some edge data here'})
        {'data_key_1': 'some edge data here'}
        >>> g.edge
        {0: {1: {'data_key_1': 'some edge data here'}}}
        """
        if type(head) is not int:
            raise ValueError("Edge head is not an integer: %s." % head)
        if type(tail) is not int:
            raise ValueError("Edge tail is not an integer: %s." % tail)
        if head < 0 or head >= len(self.node):
            raise ValueError("Head index of edge not in graph: %s." % head)
        if tail < 0 or tail >= len(self.node):
            raise ValueError("Tail index of edge not in graph: %s." % tail)
        if type(edge_data) is not dict:
            raise ValueError("Edge data is not a dict: %s." % edge_data)

        if head not in self.edge:
            self.edge[head] = {}

        self.edge[head][tail] = edge_data

        self.edge_list.append(tuple([head, tail]))

        return self.edge[head][tail]

    def add_node(self, node_data={}):
        """Create node and return (`int`, `dict`) of node key and object.

        Parameters
        ----------
        node_data: `dict`, default {}
            Data to be stored in created node

        Returns
        -------
        `tuple` of (`int`, `dict`)
            Index of created node and its data

        Raises
        ------
        ValueError
            ``node_data`` is not a ``dict``

        Examples
        --------
        >>> g = DirectedGraph()
        >>> g.add_node()
        (0, {})
        >>> g.add_node({'datakey1': 'data value'})
        (1, {'datakey1': 'data value'})
        >>> g.node
        [{}, {'datakey1': 'data value'}]
        """
        if type(node_data) is not dict:
            raise ValueError("node_data must be a dict: %s" % node_data)
        node_key = len(self.node)
        self.node.append(node_data)
        return node_key, self.node[node_key]

    def to_dict(self):
        """Convert graph to a dict, e.g. for serialize.

        Returns
        -------
        dict
            Serialization of graph as a dictionary keyed by:

            ``"edge"``
               Edge data
               (`dict` of {`int`: `dict` of {`int`: `dict`}})

            ``"node"``
                Node data
                (`list` of `dict`)

            ``"edge_list"``
                List of edges
                (`list` of `tuple` (`int`, `int`))

        Examples
        --------
        >>> g = DirectedGraph()
        >>> g.add_node()
        (0, {})
        >>> g.add_node({'datakey1': 'data value'})
        (1, {'datakey1': 'data value'})
        >>> g.node
        [{}, {'datakey1': 'data value'}]
        >>> g.add_edge(0, 1)
        {}
        >>> g.add_edge(1, 0)
        {}
        >>> g.to_dict()
        {'edge': {0: {1: {}}, 1: {0: {}}}, 'node': [{}, {'datakey1': 'data
        value'}], 'edge_list': [(0, 1), (1, 0)]}

        """

        return {"edge": self.edge, "node": self.node, "edge_list": self.edge_list}



# -*- coding: utf-8 -*-
"""Methods and schemas to validate GraphTransliterator parameters."""

from cerberus import Validator
from .process import RULE_RE, ONMATCH_RE

# ---------- validate easy reading settings ----------

EASYREADING_SETTINGS_SCHEMA = {
    "rules": {
        "type": "dict",
        "required": True,
        "keysrules": {"type": "string", "regex": RULE_RE.pattern},
        "valuesrules": {"type": "string"},
    },
    "onmatch_rules": {
        "type": "list",
        "required": False,
        "schema": {
            "type": "dict",
            "minlength": 1,
            "maxlength": 1,
            "keysrules": {"type": "string", "regex": ONMATCH_RE.pattern},
            "valuesrules": {"type": "string"},
        },
    },
    "tokens": {
        "type": "dict",
        "required": True,
        "keysrules": {"type": "string"},
        "valuesrules": {"type": "list", "schema": {"type": "string"}},
    },
    "whitespace": {
        "type": "dict",
        "required": True,
        "schema": {
            "default": {"required": True, "type": "string"},
            "token_class": {"required": True, "type": "string"},
            "consolidate": {"required": True, "type": "boolean"},
        },
    },
    "metadata": {"type": "dict", "required": False},
}


def validate_easyreading_settings(settings):
    """
    Validate general structure of "easy reading" settings, e.g.
    structure of YAML

    Parameters
    ----------
    settings : dict
        Easy reading settings to validate
    Raises
    ------
    ValueError
        If the easy reading settings do not validate.
    """
    validator = Validator()
    validator.validate(settings, EASYREADING_SETTINGS_SCHEMA, normalize=False)
    if validator.errors:
        raise ValueError(
            "Errors in GraphTransliterator settings: \n %s" % validator.errors
        )


# ---------- validate settings ----------


def validate_settings(tokens, rules, onmatch_rules, whitespace, metadata):
    """
    Validate processed settings for GraphTransliterator.

    This method first checks the tokens, then uses the tokens and token
    classes to validate the rules, onmatch rules, whitespace settings,
    all of which need the information from the tokens. Then it checks the
    metadata settings are a dictionary.

    Parameters
    ----------
    tokens: dict of {str : list of str}
        Tokens and their classes
    rules: list of dict {str: str}
        Transliteration rules in easy reading format
    onmatch_rules: list of dict {str: str}
        Rules for productions between matches of token classes
    whitespace: {'default':str, 'token_class':str, 'consolidate':bool}
        Whitespace parameters
    metadata: dict {str: str}
        Metadata settings
    Raises
    ------
    ValueError
        If there are validations errors in the "easy reading" settings
    """
    validator = Validator()

    tokens_schema = {
        "tokens": {
            "keysrules": {"type": "string"},
            "type": "dict",
            "valuesrules": {"schema": {"type": "string"}, "type": "list"},
        }
    }

    validator.validate({"tokens": tokens}, tokens_schema)

    if validator.errors:
        raise ValueError(
            "GraphTransliterator `tokens` contains invalid entries:\n %s"
            % validator.errors
        )

    token_keys = list(tokens.keys())
    token_classes = list(set().union(*tokens.values()))

    rules_schema = {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "tokens": {"required": True, "type": "list", "allowed": token_keys},
                "prev_classes": {
                    "required": False,
                    "type": "list",
                    "allowed": token_classes,
                },
                "prev_tokens": {
                    "required": False,
                    "type": "list",
                    "allowed": token_keys,
                },
                "next_tokens": {
                    "required": False,
                    "type": "list",
                    "allowed": token_keys,
                },
                "next_classes": {
                    "required": False,
                    "type": "list",
                    "allowed": token_classes,
                },
                "production": {"required": True, "type": "string"},
            },
        },
    }

    onmatch_rules_schema = {
        "type": "list",
        "required": False,
        "schema": {
            "type": "dict",
            "schema": {
                "prev_classes": {"type": "list", "schema": {"allowed": token_classes}},
                "production": {"type": "string"},
                "next_classes": {"type": "list", "schema": {"allowed": token_classes}},
            },
        },
    }

    whitespace_schema = {
        "type": "dict",
        "required": True,
        "schema": {
            "default": {"type": "string", "allowed": token_keys},
            "token_class": {"type": "string", "allowed": token_classes},
            "consolidate": {"type": "boolean"},
        },
    }

    metadata_schema = {"type": "dict", "required": False}

    schemas = {
        "whitespace": whitespace_schema,
        "onmatch_rules": onmatch_rules_schema,
        "rules": rules_schema,
        "metadata": metadata_schema,
    }

    document = {
        "whitespace": whitespace,
        "rules": rules,
        "onmatch_rules": onmatch_rules,
        "metadata": metadata,
    }  # Cerberus needs a dict

    validator.validate(document, schemas)

    if validator.errors:
        raise ValueError(
            "GraphTransliterator settings contain invalid entries:\n%s"
            % validator.errors
        )



# -*- coding: utf-8 -*-

"""
GraphTransliterator core classes.
"""

from collections import deque
import itertools
import logging
import pkg_resources
import re
import unicodedata
import yaml
from .validate import validate_easyreading_settings, validate_settings
from .process import _process_easyreading_settings
from .initialize import (
    _graph_from,
    _onmatch_rule_of,
    _onmatch_rules_lookup,
    _tokenizer_from,
    _tokens_by_class_of,
    _tokens_of,
    _transliteration_rule_of,
    _whitespace_of,
)
from .exceptions import (
    AmbiguousTransliterationRulesException,
    NoMatchingTransliterationRuleException,
    UnrecognizableInputTokenException,
)

logger = logging.getLogger("graphtransliterator")


class GraphTransliterator:
    """
    A graph-based transliteration tool that lets you convert the symbols
    of one language or script to those of another using rules that you define.

    Transliteration of tokens of an input string to an output string is
    configured by: a set of input token types with classes, pattern-match rules
    involving sequences of tokens as well as preceding or following tokens and
    token classes, insertion rules between matches, and optional consolidation
    of whitespace. Rules are ordered by specificity.

    Note
    ----
    This constructor should typically not be called directly unless
    constructing a GraphTransliterator programmatically. Use
    :meth:`from_dict`, :meth:`from_yaml`, or :meth:`from_yaml_file` instead
    for "easy reading" support.  Keyword parameters used here (
    ``check_settings``, ``check_ambiguity``, ``ignore_errors``) can be passed
    from those other constructors.

    Parameters
    ----------
    tokens : `dict` of {`str`: `list` of `str`}
        Mapping of input token types to token classes
    rules : `list` of `dict`
        `list` of dictionaries of transliteration rules with keys:

            ``"prev_classes"``
              Token classes to be matched before previous tokens and tokens
              (`list` of `str`, or `None`)

            ``"prev_tokens"``
                Specific tokens to be matched before tokens
                (`list` of `str`, or `None`)

            ``"tokens"``
                Tokens to be matched
                (`list` of `str`)

            ``"next_tokens"``
                Specific tokens to follow matched tokens
                (`list` of `str`, or `None`)

            ``"next_classes"``
                Token classes to follow tokens and next tokens
                (`list` of `str`, or `None`)

    onmatch_rules : `list` of `dict`, or `None`
        Rules for output to be inserted between tokens
        of certain classes when a transliteration rule has been matched
        but before its production string has been added to the output,
        consisting of a list of dictionaries with the keys:

            ``"prev_classes"``
                Tokens classes to be found right before start of
                match with transliteration rule
                (`list` of `str`, or `None`)

            ``"next_classes"``
                Token classes to be found from start of transliteration rule
                (`list` of `str`, or `None`)

            ``"production"``
                String to be added to output before transliteration rule
                production
                (`str`)

    whitespace: `dict`
        Whitespace settings as dictionary with the keys:

            ``"default"``
                Default whitespace token
                (`str`)

            ``"token_class"``
                Comon class of whitespace tokens
                (`str`)

            ``"consolidate"``
                If true (default), sequential whitespace characters will be
                consolidated and treated as one token, and any whitespace
                at the start or end of the input will be removed.
                (`bool`)

    metadata: `dict` or `None`
        Metadata settings
    check_settings: `bool`, optional
        If true (default), the input settings are validated.
    check_ambiguity: `bool`, optional
        If true (default), transliteration rules are checked for ambiguity.
    ignore_errors: `bool`, optional
        If true, transliteration errors are ignored and do not raise an
        exception. The default is false.

    Example
    -------
    >>> from graphtransliterator import GraphTransliterator
    >>> settings = {
    ...    'tokens': {'a': ['class1'],
    ...               'b': ['class2'],
    ...               ' ': ['wb']},
    ...    'rules': [
    ...        {'production': 'A', 'tokens': ['a']},
    ...        {'production': 'B', 'tokens': ['b']},
    ...        {'production': ' ', 'tokens': [' ']},
    ...        {'production': 'A*',
    ...         'prev_classes': ['class2'],
    ...         'prev_tokens': ['a'],
    ...         'tokens': ['a'],
    ...         'next_tokens': ['a'],
    ...         'next_classes': ['class2'] }],
    ...    'onmatch_rules': [
    ...        {'prev_classes': ['class1'],
    ...         'next_classes': ['class1'],
    ...         'production': ','}],
    ...    'whitespace': {
    ...        'default': ' ',
    ...        'consolidate': False,
    ...        'token_class': 'wb'},
    ...    'metadata': {
    ...        'author': 'Author McAuthorson',
    ...        'version': '0.0.1'}
    ... }
    >>> gt = GraphTransliterator(
    ...         settings['tokens'], settings['rules'], settings['onmatch_rules'],
    ...         settings['whitespace'], settings['metadata']
    ...      )
    >>> gt.transliterate('baaab')
    'BA,A*,AB'
    >>>

    See Also
    --------
    from_dict : constructor from  dictionary in "easy reading" format
    from_yaml : constructor from YAML string in "easy reading" format
    from_yaml_file : constructor from YAML in "easy reading" format
"""  # noqa

    def __init__(
        self,
        tokens,
        rules,
        onmatch_rules,
        whitespace,
        metadata,
        check_settings=True,
        check_ambiguity=True,
        ignore_errors=False,
        **kwargs
    ):
        if check_settings:
            validate_settings(tokens, rules, onmatch_rules, whitespace, metadata)
        self._metadata = metadata
        self._tokens = _tokens_of(tokens)
        self._tokens_by_class = _tokens_by_class_of(tokens)
        self._rules = sorted(
            [_transliteration_rule_of(rule) for rule in rules],
            key=lambda transliteration_rule: transliteration_rule.cost,
        )
        if check_ambiguity:
            self.check_for_ambiguity()
        self._onmatch_rules = [_onmatch_rule_of(_) for _ in onmatch_rules]
        self._onmatch_rules_lookup = _onmatch_rules_lookup(tokens, self._onmatch_rules)
        self._whitespace = _whitespace_of(whitespace)
        self._ignore_errors = ignore_errors
        self.__init_parameters__ = {  # used by pruned_of
            "tokens": tokens,
            "rules": rules,
            "onmatch_rules": onmatch_rules,
            "whitespace": whitespace,
            "metadata": metadata,
            "check_settings": check_settings,
            "ignore_errors": ignore_errors,
            "check_ambiguity": check_ambiguity,
            "kwargs": kwargs,
        }
        self._tokenizer = _tokenizer_from(list(tokens.keys()))
        self._graph = _graph_from(self.rules)
        # self._input_tokens = None
        self._rule_keys = []  # last matched rules

    def _match_constraints(self, source, target, token_i, tokens):
        """
        Match edge constraints.

        Called on edge before a rule. `token_i` is set to location right
        after tokens consumed.

        """
        target_edge = self._graph.edge[source][target]

        constraints = target_edge.get("constraints")

        if not constraints:
            return True
        # should these be ordered?
        for c_type, c_value in constraints.items():
            if c_type == "prev_tokens":
                num_tokens = len(self._graph.node[target]["rule"].tokens)
                # presume for rule (a) a, with input "aa"
                # ' ', a, a, ' '  start (token_i=3)
                #             ^
                #         ^       -1 subtract num_tokens
                #      ^          - len(c_value)
                start_at = token_i
                start_at -= num_tokens
                start_at -= len(c_value)

                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=True,
                    check_next=False,
                    by_class=False,
                ):
                    return False
            elif c_type == "next_tokens":
                # presume for rule a (a), with input "aa"
                # ' ', a, a, ' '  start (token_i=2)
                #         ^
                start_at = token_i

                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=False,
                    check_next=True,
                    by_class=False,
                ):
                    return False

            elif c_type == "prev_classes":
                num_tokens = len(self._graph.node[target]["rule"].tokens)
                # presume for rule (a <class_a>) a, with input "aaa"
                # ' ', a, a, a, ' '
                #                ^     start (token_i=4)
                #            ^         -num_tokens
                #         ^            -len(prev_tokens)
                #  ^                   -len(prev_classes)
                start_at = token_i
                start_at -= num_tokens
                prev_tokens = constraints.get("prev_tokens")
                if prev_tokens:
                    start_at -= len(prev_tokens)
                start_at -= len(c_value)
                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=True,
                    check_next=False,
                    by_class=True,
                ):
                    return False

            elif c_type == "next_classes":
                # presume for rule a (a <class_a>), with input "aaa"
                # ' ', a, a, a, ' '
                #         ^          start (token_i=2)
                #            ^       + len of next_tokens (a)
                start_at = token_i
                next_tokens = constraints.get("next_tokens")
                if next_tokens:
                    start_at += len(next_tokens)
                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=False,
                    check_next=True,
                    by_class=True,
                ):
                    return False

        return True

    def match_at(self, token_i, tokens, match_all=False):
        """
        Match best (least costly) transliteration rule at a given index in the
        input tokens and return the index to  that rule. Optionally, return all
        rules that match.

        Parameters
        ----------
        token_i : `int`
            Location in `tokens` at which to begin
        tokens : `list` of `str`
            List of tokens
        match_all : `bool`, optional
            If true, return the index of all rules matching at the given
            index. The default is false.

        Returns
        -------
        `int`, `None`, or `list` of `int`
            Index of matching transliteration rule in
            :attr:`GraphTransliterator.rules` or None. Returns a `list` of
            `int` or an empty `list` if ``match_all`` is true.

        Note
        ----
        Expects whitespaces token at beginning and end of `tokens`.

        Examples
        --------

        >>> from graphtransliterator import GraphTransliterator
        >>> gt = GraphTransliterator.from_yaml('''
        ...         tokens:
        ...             a: []
        ...             a a: []
        ...             ' ': [wb]
        ...         rules:
        ...             a: <A>
        ...             a a: <AA>
        ...         whitespace:
        ...             default: ' '
        ...             consolidate: True
        ...             token_class: wb
        ... ''')
        >>> tokens = gt.tokenize("aa")
        >>> tokens # whitespace added to ends
        [' ', 'a', 'a', ' ']
        >>> gt.match_at(1, tokens) # returns index to rule
        0
        >>> gt.rules[gt.match_at(1, tokens)] # actual rule
        TransliterationRule(production='<AA>', prev_classes=None, prev_tokens=None, tokens=['a', 'a'], next_tokens=None, next_classes=None, cost=0.41503749927884376)
        >>> gt.match_at(1, tokens, match_all=True) # index to rules, with match_all
        [0, 1]
        >>>
        >>> [gt.rules[_] for _ in gt.match_at(1, tokens, match_all=True)]
        [TransliterationRule(production='<AA>', prev_classes=None, prev_tokens=None, tokens=['a', 'a'], next_tokens=None, next_classes=None, cost=0.41503749927884376), TransliterationRule(production='<A>', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562)]
        >>>
        """  # noqa

        graph = self._graph
        if match_all:
            matches = []
        stack = deque()

        def _append_children(node_key, token_i):
            children = None
            ordered_children = graph.node[node_key].get("ordered_children")
            if ordered_children:
                children = ordered_children.get(tokens[token_i])
                if children:
                    # reordered high to low for stack:
                    for child_key in reversed(children):
                        stack.appendleft((child_key, node_key, token_i))
                else:
                    rules_keys = ordered_children.get("__rules__")  # leafs
                    if rules_keys:
                        # There may be more than one as certain rules have
                        # constraints on them.
                        # reordered so high cost go on stack last
                        for rule_key in reversed(rules_keys):
                            stack.appendleft((rule_key, node_key, token_i))

        _append_children(0, token_i)  # append all children of root node

        while stack:  # LIFO
            node_key, parent_key, token_i = stack.popleft()
            assert token_i < len(tokens), "way past boundary"
            curr_node = graph.node[node_key]
            # check constraints on preceding edge
            if curr_node.get("accepting") and self._match_constraints(
                parent_key, node_key, token_i, tokens
            ):
                if match_all:
                    matches.append(curr_node["rule_key"])
                    continue
                else:
                    return curr_node["rule_key"]
            else:
                if token_i < len(tokens) - 1:
                    token_i += 1
                _append_children(node_key, token_i)
        if match_all:
            return matches

    def _match_tokens(
        self, start_i, c_value, tokens, check_prev=True, check_next=True, by_class=False
    ):
        """Match tokens, with boundary checks."""

        if check_prev and start_i < 0:
            return False
        if check_next and start_i + len(c_value) > len(tokens):
            return False
        for i in range(0, len(c_value)):
            if by_class:
                if not c_value[i] in self._tokens[tokens[start_i + i]]:
                    return False
            elif tokens[start_i + i] != c_value[i]:
                return False
        return True

    @property
    def graph(self):
        """`DirectedGraph`: Graph used in transliteration."""
        return self._graph

    @property
    def ignore_errors(self):
        """`bool`: Ignore transliteration errors setting."""
        return self._ignore_errors

    @ignore_errors.setter
    def ignore_errors(self, value):
        self._ignore_errors = value

    @property
    def last_matched_rules(self):
        """
        `list` of `TransliterationRule`: Last transliteration rules matched.
        """
        return [self._rules[_] for _ in self._rule_keys]

    @property
    def last_matched_rule_tokens(self):
        """`list` of `list` of `str`: Last matched tokens for each rule."""
        return [self._rules[_].tokens for _ in self._rule_keys]

    @property
    def last_input_tokens(self):
        """
        `list` of `str`: Last tokenization of the input string, with whitespace
        at start and end."""
        return self._input_tokens

    @property
    def metadata(self):
        """
        `dict`: Metadata of transliterator
        """
        return self._metadata

    def transliterate(self, input):
        """
        Transliterate an input string into an output string.

        Parameters
        ----------
        input : `str`
            Input string to transliterate

        Returns
        -------
        `str`
            Transliteration output string

        Raises
        ------
        ValueError
            Cannot parse input

        Note
        ----
        Whitespace will be temporarily appended to start and end of input
        string.

        Example
        -------
        >>> from graphtransliterator import GraphTransliterator
        >>> GraphTransliterator.from_yaml(
        ... '''
        ... tokens:
        ...   a: []
        ...   ' ': [wb]
        ... rules:
        ...   a: A
        ...   ' ': '_'
        ... whitespace:
        ...   default: ' '
        ...   consolidate: True
        ...   token_class: wb
        ... ''').transliterate("a a")
        'A_A'
        """
        tokens = self.tokenize(input)  # adds initial+final whitespace
        self._input_tokens = tokens  # <--- tokens are saved here
        self._rule_keys = []
        output = ""
        token_i = 1  # adjust for initial whitespace

        while token_i < len(tokens) - 1:  # adjust for final whitespace
            rule_key = self.match_at(token_i, tokens)
            if rule_key is None:
                logger.warning(
                    "No matching transliteration rule at token pos %s of %s"
                    % (token_i, tokens)
                )
                # No parsing rule was found at this location
                if self.ignore_errors:
                    # move along if ignoring errors
                    token_i += 1
                    continue
                else:
                    raise NoMatchingTransliterationRuleException
            self._rule_keys.append(rule_key)
            rule = self.rules[rule_key]
            tokens_matched = rule.tokens
            if self._onmatch_rules:
                curr_match_rules = None
                prev_t = tokens[token_i - 1]
                curr_t = tokens[token_i]
                curr_t_rules = self._onmatch_rules_lookup.get(curr_t)
                if curr_t_rules:
                    curr_match_rules = curr_t_rules.get(prev_t)
                if curr_match_rules:
                    for onmatch_i in curr_match_rules:
                        onmatch = self._onmatch_rules[onmatch_i]
                        # <class_a> <class_a> + <class_b>
                        # a a b
                        #     ^
                        # ^      - len(onmatch.prev_rules)
                        if self._match_tokens(
                            token_i - len(onmatch.prev_classes),
                            onmatch.prev_classes,  # double checks last value
                            tokens,
                            check_prev=True,
                            check_next=False,
                            by_class=True,
                        ) and self._match_tokens(
                            token_i,
                            onmatch.next_classes,  # double checks first value
                            tokens,
                            check_prev=False,
                            check_next=True,
                            by_class=True,
                        ):
                            output += onmatch.production
                            break  # only match best onmatch rule
            output += rule.production
            token_i += len(tokens_matched)
        return output

    def tokenize(self, input):
        """
        Tokenizes an input string.

        Adds initial and trailing whitespace, which can be consolidated.

        Parameters
        ----------
        input : str
            String to tokenize

        Returns
        -------
        `list` of `str`
            List of tokens, with default whitespace token at beginning and end.

        Raises
        ------
        ValueError
            Unrecognizable input, such as a character that is not in a token

        Examples
        --------
        >>> from graphtransliterator import GraphTransliterator
        >>> t = {'ab': ['class_ab'], ' ': ['wb']}
        >>> w = {'default': ' ', 'token_class': 'wb', 'consolidate': True}
        >>> r = {'ab': 'AB', ' ': '_'}
        >>> settings = {'tokens': t, 'rules': r, 'whitespace': w}
        >>> gt = GraphTransliterator.from_dict(settings)
        >>> gt.tokenize('ab ')
        >>> [' ', 'ab', ' ']
        """

        def is_whitespace(token):
            """Check if token is whitespace."""
            return self.whitespace.token_class in self.tokens[token]

        # start with a whitespace token
        tokens = [self.whitespace.default]

        prev_whitespace = True

        match_at = 0
        while match_at < len(input):
            match = self._tokenizer.match(input, match_at)
            if match:
                match_at = match.end()  # advance match_at
                token = match.group(0)
                # Could save match loc here:
                # matched_at = match.span(0)[0]
                if is_whitespace(token):
                    if prev_whitespace and self.whitespace.consolidate:
                        continue
                    else:
                        prev_whitespace = True
                else:
                    prev_whitespace = False
                tokens.append(token)
            else:
                logger.warning(
                    "Unrecognizable token %s at pos %s of %s"
                    % (input[match_at], match_at, input)
                )
                if not self.ignore_errors:
                    raise UnrecognizableInputTokenException
                else:
                    match_at += 1

        if self.whitespace.consolidate:
            while len(tokens) > 1 and is_whitespace(tokens[-1]):
                tokens.pop()

        tokens.append(self.whitespace.default)

        assert len(tokens) >= 2  # two whitespaces, at least

        return tokens

    def pruned_of(self, productions):
        """
        Remove transliteration rules with specific output productions.

        Parameters
        ----------
        productions : `str`, or `list` of `str`
            list of productions to remove

        Returns
        -------
        graphtransliterator.GraphTransliterator
            Graph transliterator pruned of certain productions.

        Note
        ----
        Uses original initialization parameters to construct a new
        :class:`GraphTransliterator`.

        Examples
        --------
        >>> gt = GraphTransliterator.from_yaml('''
        ...         tokens:
        ...             a: []
        ...             a a: []
        ...             ' ': [wb]
        ...         rules:
        ...             a: <A>
        ...             a a: <AA>
        ...         whitespace:
        ...             default: ' '
        ...             consolidate: True
        ...             token_class: wb
        ... ''')
        >>> gt.rules
        [TransliterationRule(production='<AA>', prev_classes=None, prev_tokens=None, tokens=['a', 'a'], next_tokens=None, next_classes=None, cost=0.41503749927884376), TransliterationRule(production='<A>', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562)]
        >>> gt.pruned_of('<AA>').rules
        [TransliterationRule(production='<A>', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562)]
        >>> gt.pruned_of(['<A>', '<AA>']).rules
        []
        """  # noqa
        __init_parameters__ = self.__init_parameters__
        pruned_rules = [
            _
            for _ in __init_parameters__["rules"]
            if not _["production"] in productions
        ]
        return GraphTransliterator(
            __init_parameters__["tokens"],
            pruned_rules,
            __init_parameters__["onmatch_rules"],
            __init_parameters__["whitespace"],
            __init_parameters__["metadata"],
            ignore_errors=__init_parameters__["ignore_errors"],
            check_settings=__init_parameters__["check_settings"],
            check_ambiguity=__init_parameters__["check_ambiguity"],
            **__init_parameters__["kwargs"]
        )

    @property
    def productions(self):
        """
        `list` of `str`: List of productions of each transliteration rule.
        """
        return [_.production for _ in self._rules]

    @property
    def tokens(self):
        """
        `dict` of {`str`:`set` of `str`}: Mappings of tokens to their classes.
        """
        return self._tokens

    @property
    def rules(self):
        """
        `list` of `TransliterationRule`: Transliteration rules sorted by cost.
        """
        return self._rules

    @property
    def onmatch_rules(self):
        """`list` of `OnMatchRules`: Rules for productions between matches."""
        return self._onmatch_rules

    @property
    def whitespace(self):
        """WhiteSpaceRules: Whitespace rules."""
        return self._whitespace

    @classmethod
    def from_dict(cls, settings, **kwargs):
        """
        Constructs `GraphTransliterator` from a dictionary of settings in
        "easy reading" format, i.e. the loaded contents of a YAML string.

        Parameters
        ----------
        settings : `dict`
            Settings dictionary in easy reading format with keys:

                ``"tokens"``
                  Mappings of tokens to their classes
                  (`dict` of {str: `list` of `str`})

                ``"rules"``
                  Transliteration rules in "easy reading" format
                  (`list` of `dict` of {`str`: `str`})

                ``"onmatch_rules"``
                  On match rules in "easy reading" format
                  (`dict` of {`str`: `str`}, optional)

                ``"whitespace"``
                  Whitespace definitions, including default whitespace token,
                  class of whitespace tokens, and whether or not to consolidate
                  (`dict` of {'default': `str`, 'token_class': `str`,
                  consolidate: `bool`}, optional)

                ``"metadata"``
                  Dictionary of metadata (`dict`, optional)

        Returns
        -------
        `GraphTransliterator`

        Note
        ----
        Called by :meth:`from_yaml`.

        Example
        -------
        >>> from graphtransliterator import GraphTransliterator
        >>> tokens = {
        ...     'ab': ['class_ab'],
        ...     ' ': ['wb']
        ... }
        >>> whitespace = {
        ...     'default': ' ',
        ...     'token_class': 'wb',
        ...     'consolidate': True
        ... }
        >>> onmatch_rules = [
        ...     {'<class_ab> + <class_ab>': ','}
        ... ]
        >>> rules = {'ab': 'AB',
        ...          ' ': '_'}
        >>> settings = {'tokens': tokens,
        ...             'rules': rules,
        ...             'whitespace': whitespace,
        ...             'onmatch_rules': onmatch_rules}
        >>> gt = GraphTransliterator.from_dict(settings)
        >>> gt.transliterate("ab abab")
        'AB_AB,AB'
        >>>

        See Also
        --------
        GraphTransliterator.yaml()
        GraphTransliterator.from_yaml_file()
        """
        validate_easyreading_settings(settings)
        settings = _process_easyreading_settings(settings)

        return GraphTransliterator(
            settings["tokens"],
            settings["rules"],
            settings["onmatch_rules"],
            settings["whitespace"],
            settings["metadata"],
            **kwargs
        )

    @classmethod
    def from_yaml(cls, yaml_str, charnames_escaped=True, **kwargs):
        """
        Construct GraphTransliterator from a YAML str.

        Parameters
        ----------
        yaml_str : str
            YAML mappings of tokens, rules, and (optionally) onmatch_rules
        charnames_escaped : boolean
            Unescape Unicode during YAML read (default True)

        Note
        ----
        Called by :meth:`from_yaml_file` and calls :meth:`from_dict`.

        Example
        -------
        >>> from graphtransliterator import GraphTransliterator
        >>> yaml_ = '''
        ... tokens:
        ...   a: [class1]
        ...   ' ': [wb]
        ... rules:
        ...   a: A
        ...   ' ': ' '
        ... whitespace:
        ...   default: ' '
        ...   consolidate: True
        ...   token_class: wb
        ... onmatch_rules:
        ...   - <class1> + <class1>: "+"
        ... '''
        >>> gt = GraphTransliterator.from_yaml(yaml_)
        >>> gt.transliterate("a aa")
        'A A+A'

        See Also
        --------
        GraphTransliterator.from_dict()
        GraphTransliterator.from_yaml_file()
        """
        if charnames_escaped:
            yaml_str = _unescape_charnames(yaml_str)

        settings = yaml.safe_load(yaml_str)

        return cls.from_dict(settings, **kwargs)

    @classmethod
    def from_yaml_file(cls, yaml_filename, **kwargs):
        """
        Construct GraphTransliterator from YAML file.

        Parameters
        ----------
        yaml_filename : str
            Name of YAML file, containing tokens, rules, and (optionally)
            onmatch_rules

        Note
        ----
        Calls :meth:`from_yaml`.

        See Also
        --------
        graphtransliterator.GraphTransliterator.from_yaml
        graphtransliterator.GraphTransliterator.from_dict
        """
        with open(yaml_filename, "r") as f:
            yaml_string = f.read()

        return cls.from_yaml(yaml_string, **kwargs)

    def serialize(self):
        """
        Serialize all relevant settings of Graph Transliterator.

        Returns
        -------
        dict
            GraphTransliterator configuration as a dictionary with keys:

                ``"graph"``
                  Serialization of `DirectedGraph`
                  (`dict`)

                ``"tokenizer_pattern"``
                  Regular expression for tokenizing
                  (`str`)

                ``"tokens"``
                  Mappings of tokens to their classes
                  (`dict` of {str: `list` of `str`})

                ``"rules"``
                  Transliteration rules in direct format
                  (`list` of `dict` of {`str`: `str`})

                ``"onmatch_rules"``
                  Onmatch settings
                  (`list` of `OnMatchRule`)

                ``"onmatch_rules_lookup"``
                  Dictionary keyed by current token to previous token
                  containing a list of indexes of applicable `OnmatchRule`
                  to try
                  (`dict` of {`str`: `dict` of {`str`: `list` of `int`}})

                ``"whitespace"``
                  Whitespace settings
                  (`WhitespaceRules`)

                ``"metadata"``
                  Dictionary of metadata (`dict`)

                ``"graphtransliterator_version"``
                  Module version of `graphtransliterator` (`str`)

        Note
        ----
        `OnmatchRule`, `TransliterationRule`, and `WhitespaceRules` are
        initialized from :meth:`collections.namedtuple`. Therefore, their keys
        will always be in the same order and can be accessed by index if
        reading from a non-Python format.

        Example
        -------
        >>> from graphtransliterator import GraphTransliterator
        >>> yaml_ = '''
        ...  tokens:
        ...    a: [vowel]
        ...    ' ': [wb]
        ...  rules:
        ...    a: A
        ...    ' ': ' '
        ...  whitespace:
        ...    default: " "
        ...    consolidate: false
        ...    token_class: wb
        ...  onmatch_rules:
        ...    - <vowel> + <vowel>: ','  # add a comma between vowels
        ... '''
        >>> gt = GraphTransliterator.from_yaml(yaml_)
        >>> import pprint
        >>> pprint.pprint(gt.serialize())
        {'graph': {'edge': {0: {1: {'cost': 0.5849625007211562, 'token': 'a'},
                                3: {'cost': 0.5849625007211562, 'token': ' '}},
                            1: {2: {'cost': 0.5849625007211562}},
                            3: {4: {'cost': 0.5849625007211562}}},
                   'edge_list': [(0, 1), (1, 2), (0, 3), (3, 4)],
                   'node': [{'ordered_children': {' ': [3], 'a': [1]}, 'type': 'Start'},
                            {'ordered_children': {'__rules__': [2]},
                             'token': 'a',
                             'type': 'token'},
                            {'accepting': True,
                             'ordered_children': {},
                             'rule': TransliterationRule(production='A', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562),
                             'rule_key': 0,
                             'type': 'rule'},
                            {'ordered_children': {'__rules__': [4]},
                             'token': ' ',
                             'type': 'token'},
                            {'accepting': True,
                             'ordered_children': {},
                             'rule': TransliterationRule(production=' ', prev_classes=None, prev_tokens=None, tokens=[' '], next_tokens=None, next_classes=None, cost=0.5849625007211562),
                             'rule_key': 1,
                             'type': 'rule'}]},
         'graphtransliterator_version': '0.2.5',
         'metadata': {},
         'onmatch_rules': [OnMatchRule(prev_classes=['vowel'], next_classes=['vowel'], production=',')],
         'onmatch_rules_lookup': {'a': {'a': [0]}},
         'rules': [TransliterationRule(production='A', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562),
                   TransliterationRule(production=' ', prev_classes=None, prev_tokens=None, tokens=[' '], next_tokens=None, next_classes=None, cost=0.5849625007211562)],
         'tokenizer_pattern': '(a|\\ )',
         'tokens': {' ': ['wb'], 'a': ['vowel']},
         'whitespace': WhitespaceRules(default=' ', token_class='wb', consolidate=False)}
        """  # noqa

        version = pkg_resources.require("graphtransliterator")[0].version
        return {
            "graph": self._graph.to_dict(),
            "metadata": self._metadata,
            "tokens": {token: list(classes) for token, classes in self._tokens.items()},
            "rules": self._rules,
            "onmatch_rules": self._onmatch_rules,
            "onmatch_rules_lookup": self._onmatch_rules_lookup,
            "whitespace": self._whitespace,
            "tokenizer_pattern": self._tokenizer.pattern,
            "graphtransliterator_version": version,
        }

    def check_for_ambiguity(self):
        """
        Check if multiple transliteration rules could match the same tokens.

        This function first groups the transliteration rules by number of
        tokens. It then checks to see if any pair of the same cost would match
        the same sequence of tokens. If so, it finally checks if a less costly
        rule would match those particular sequences. If not, there is
        ambiguity.

        Details of all ambiguity are sent in a :func:`logging.warning`.

        Note
        ----
        Called during initialization if ``check_ambiguity`` is set.

        Raises
        ------
        AmbiguousTransliterationRulesException
            Multiple transliteration rules could match the same tokens.

        Example
        -------
        >>> from graphtransliterator import GraphTransliterator
        >>> yaml_filename = '''
        ... tokens:
        ...   a: [class1, class2]
        ...   ' ': [wb]
        ... rules:
        ...   <class1> a: AW
        ...   <class2> a: AA # ambiguous rule
        ... whitespace:
        ...   default: ' '
        ...   consolidate: True
        ...   token_class: wb
        ... '''
        >>> gt = GraphTransliterator.from_yaml(yaml_, check_ambiguity=False)
        >>> gt.check_for_ambiguity()
        WARNING:root:The pattern [{'a'}, {'a'}] can be matched by both:
          <class1> a
          <class2> a
        ...
        graphtransliterator.exceptions.AmbiguousTransliterationRulesException
        >>>
        """

        ambiguity = False

        all_tokens = set(self.tokens.keys())

        rules = self._rules

        if not rules:
            return True

        max_prev = [_count_of_prev(rule) for rule in rules]
        global_max_prev = max(max_prev)
        max_curr_next = [_count_of_curr_and_next(rule) for rule in rules]
        global_max_curr_next = max(max_curr_next)

        # Generate a matrix of rules, where width is the max of
        # any previous tokens/classes + max of current/next tokens/classes.
        # Each rule's specifications starting from the max of the previous
        # tokens/classes. Other positions are filled by the set of all possible
        # tokens.

        matrix = []

        width = global_max_prev + global_max_curr_next

        for i, rule in enumerate(rules):
            row = [all_tokens] * (global_max_prev - max_prev[i])
            row += _tokens_possible(rule, self._tokens_by_class)
            row += [all_tokens] * (width - len(row))
            matrix += [row]

        def full_intersection(i, j):
            """ Intersection of  matrix[i] and matrix[j], else None."""

            intersections = []
            for k in range(width):
                intersection = matrix[i][k].intersection(matrix[j][k])
                if not intersection:
                    return None
                intersections += [intersection]
            return intersections

        def covered_by(intersection, row):
            """Check if intersection is covered by row."""
            for i in range(len(intersection)):
                diff = intersection[i].difference(row[i])
                if diff:
                    return False
            return True

        # Iterate through rules based on number of tokens (cost).
        # If there are ambiguities, then see if a less costly rule
        # would match the rule. If it does not, there is ambiguity.

        grouper = lambda x: (_count_of_tokens(x))  # noqa, could replace by cost

        for group_val, group_iter in itertools.groupby(
            enumerate(self._rules), key=lambda x: grouper(x[1])
        ):

            group = list(group_iter)
            if len(group) == 1:
                continue
            for i in range(len(group) - 1):
                for j in range(i + 1, len(group)):
                    i_index = group[i][0]
                    j_index = group[j][0]
                    intersection = full_intersection(i_index, j_index)
                    if not intersection:
                        break

                    # check if a less costly rule matches intersection

                    def covered_by_less_costly():
                        for r_i, rule in enumerate(rules):
                            if r_i in (i_index, j_index):
                                continue
                            if rule.cost > rules[i_index].cost:
                                continue
                            rule_tokens = matrix[r_i]
                            if covered_by(intersection, rule_tokens):
                                return True
                        return False

                    if not covered_by_less_costly():
                        logging.warning(
                            "The pattern {} can be matched by both:\n"
                            "  {}\n"
                            "  {}\n".format(
                                intersection,
                                _easyreading_rule(rules[i_index]),
                                _easyreading_rule(rules[j_index]),
                            )
                        )
                        ambiguity = True
        if ambiguity:
            raise AmbiguousTransliterationRulesException


def _easyreading_rule(rule):
    """Get an easy-reading string of a rule."""

    def _token_str(x):
        return " ".join(x)

    def _class_str(x):
        return " ".join(["<%s>" % _ for _ in x])

    out = ""
    if rule.prev_classes and rule.prev_tokens:
        out = "({} {}) ".format(
            _class_str(rule.prev_classes), _token_str(rule.prev_tokens)
        )
    elif rule.prev_classes:
        out = "{} ".format(_class_str(rule.prev_classes))
    elif rule.prev_tokens:
        out = "({}) ".format(_token_str(rule.prev_tokens))

    out += _token_str(rule.tokens)

    if rule.next_tokens and rule.next_classes:
        out += " ({} {})".format(
            _token_str(rule.next_tokens), _class_str(rule.next_classes)
        )
    elif rule.next_tokens:
        out += " ({})".format(_token_str(rule.next_tokens))
    elif rule.next_classes:
        out += " {}".format(_class_str(rule.next_classes))
    return out


def _count_of_prev(rule):
    """Count previous tokens to be present before a match in a rule."""

    return len(rule.prev_classes or []) + len(rule.prev_tokens or [])


def _count_of_curr_and_next(rule):
    """Count tokens to be matched and those to follow them in rule."""

    return len(rule.tokens) + len(rule.next_tokens or []) + len(rule.next_classes or [])


def _count_of_tokens(rule):
    return (
        len(rule.prev_classes or [])
        + len(rule.prev_tokens or [])
        + len(rule.tokens)
        + len(rule.next_tokens or [])
        + len(rule.next_classes or [])
    )


def _prev_tokens_possible(rule, tokens_by_class):
    """`list` of set of possible preceding tokens for a rule."""

    return [tokens_by_class[_] for _ in rule.prev_classes or []] + [
        set([_]) for _ in rule.prev_tokens or []
    ]


def _curr_and_next_tokens_possible(rule, tokens_by_class):
    """`list` of sets of possible current and following tokens for a rule."""

    return (
        [set([_]) for _ in rule.tokens]
        + [set([_]) for _ in rule.next_tokens or []]
        + [tokens_by_class[_] for _ in rule.next_classes or []]
    )


def _tokens_possible(row, tokens_by_class):
    """`list` of sets of possible tokens matched for a rule."""

    return _prev_tokens_possible(row, tokens_by_class) + _curr_and_next_tokens_possible(
        row, tokens_by_class
    )


# initialization-related functions for unescaping unicode


def _unescape_charnames(input_str):
    r"""
    Convert \\N{Unicode charname}-escaped str to unicode characters.

    This is useful for specifying exact character names, and a default
    escape feature in Python that needs a function to be used for reading
    from files.

    Parameters
    ----------
    input_str : str
        The unescaped string, with \\N{Unicode charname} converted to
        the corresponding Unicode characters.

    Examples
    --------

    >>> from graphtransliterator import GraphTransliterator
    >>> GraphTransliterator._unescape_charnames(r"H\N{LATIN SMALL LETTER I}")
    'Hi'
    """

    def get_unicode_char(matchobj):
        """Get Unicode character value from escaped character sequences."""
        charname = matchobj.group(0)
        match = re.match(r"\\N{([A-Z ]+)}", charname)
        char = unicodedata.lookup(match.group(1))  # KeyError if invalid
        return char

    return re.sub(r"\\N{[A-Z ]+}", get_unicode_char, input_str)



# -*- coding: utf-8 -*-

"""
Functions used to initialize a GraphTransliterator.
"""

from collections import defaultdict
from .graphs import DirectedGraph
import math
import re
from .rules import TransliterationRule, WhitespaceRules, OnMatchRule

# ---------- initialize tokens ----------


def _tokens_of(tokens):
    """Converts values of dict of tokens from list to set."""

    return {key: set(value) for key, value in tokens.items()}


def _tokens_by_class_of(tokens):
    """Generates lookup table of tokens in each class."""

    out = defaultdict(set)
    for token, token_classes in tokens.items():
        for token_class in token_classes:
            out[token_class].add(token)
    return out


# ---------- initialize rules ----------


def _transliteration_rule_of(rule):
    """Converts rule dict into a TransliterationRule."""

    _cost = _cost_of(rule)

    transliteration_rule = TransliterationRule(
        production=rule.get("production"),
        prev_classes=rule.get("prev_classes"),
        prev_tokens=rule.get("prev_tokens"),
        tokens=rule.get("tokens"),
        next_tokens=rule.get("next_tokens"),
        next_classes=rule.get("next_classes"),
        cost=_cost,
    )

    return transliteration_rule


def _num_tokens_of(rule):
    """Calculate the total number of tokens in a rule."""

    total = (
        len(rule.get("prev_classes", []))
        + len(rule.get("prev_tokens", []))
        + len(rule.get("tokens"))
        + len(rule.get("next_tokens", []))
        + len(rule.get("next_classes", []))
    )

    return total


def _cost_of(rule):
    """Calculate the cost of a rule based on the number of constraints.

    Rules requiring more tokens to match are made less costly and tried first.
    """

    return math.log2(1 + 1 / (1 + _num_tokens_of(rule)))


def _onmatch_rule_of(onmatch_rule):
    """ Converts onmatch_rule into OnMatchRule. """

    return OnMatchRule(
        prev_classes=onmatch_rule["prev_classes"],
        next_classes=onmatch_rule["next_classes"],
        production=onmatch_rule["production"],
    )


def _onmatch_rules_lookup(tokens, onmatch_rules):
    """ Creates a dict lookup from current to previous token.

    Returns
    -------
    dict of {str: dict of {str: list of int}}
        Dictionary keyed by current token to previous token containing a
        list of onmatch rules in order that would apply
    """

    onmatch_lookup = {}
    # for token_key in tokens:
    #     onmatch_lookup[token_key] = {_: [] for _ in tokens}

    # Interate through onmatch rules
    for rule_i, rule in enumerate(onmatch_rules):
        # Iterate through all tokens
        for token_key, token_classes in tokens.items():
            # if the onmatch rule's next is of that class
            if rule.next_classes[0] in token_classes:
                # iterate through all tokens again
                for prev_token_key, prev_token_classes in tokens.items():
                    # if second token is of class of onmatch rule's last prev
                    # add the rule to curr -> prev
                    if rule.prev_classes[-1] in prev_token_classes:
                        if token_key not in onmatch_lookup:
                            onmatch_lookup[token_key] = {}
                        curr_token = onmatch_lookup[token_key]
                        if prev_token_key not in curr_token:
                            curr_token[prev_token_key] = []
                        rule_list = curr_token[prev_token_key]
                        rule_list.append(rule_i)
    return onmatch_lookup


# ---------- initialize whitespace ---------


def _whitespace_of(whitespace):
    """ Converts whitespace into WhiteSpace """
    return WhitespaceRules(
        default=whitespace["default"],
        token_class=whitespace["token_class"],
        consolidate=whitespace["consolidate"],
    )


# ---------- initialize tokenizer ----------


def _tokenizer_from(tokens):
    """ Generates regular expression tokenizer from token list."""

    tokens.sort(key=len, reverse=True)
    regex_str = "(" + "|".join([re.escape(_) for _ in tokens]) + ")"
    return re.compile(regex_str, re.S)


# ---------- initialize graph ----------


def _graph_from(rules):
    """Generates a parsing graph from the given rules.

    The directed graph is a tree, and the rules are its leaves. Each
    intermediate node represents a token. Before trying to match using a
    token, the constraints on the preceding edge must be met. These include
    token and, on the edge before  a rule, previous and next tokens and token
    classes. The tree is built top-down. Costs of edges are adjusted to match
    the lowest cost of leaf nodes, so they are tried first."""

    graph = DirectedGraph()
    graph.add_node({"type": "Start"})

    for rule_key, rule in enumerate(rules):
        parent_key = 0
        for token in rule.tokens:
            parent_node = graph.node[parent_key]
            token_children = parent_node.get("token_children") or {}
            token_node_key = token_children.get(token)
            if token_node_key is None:
                token_node_key, _ = graph.add_node({"type": "token", "token": token})
                graph.add_edge(parent_key, token_node_key, {"token": token})
                token_children[token] = token_node_key
                parent_node["token_children"] = token_children

            curr_edge = graph.edge[parent_key][token_node_key]
            curr_cost = curr_edge.get("cost", 1)
            if curr_cost > rule.cost:
                curr_edge["cost"] = rule.cost

            parent_key = token_node_key

        rule_node_key, _ = graph.add_node(
            {"type": "rule", "rule_key": rule_key, "rule": rule, "accepting": True}
        )
        parent_node = graph.node[parent_key]
        rule_children = parent_node.get("rule_children", [])
        rule_children.append(rule_node_key)
        parent_node["rule_children"] = sorted(
            rule_children, key=lambda x: graph.node[x]["rule"].cost
        )

        edge_to_rule = graph.add_edge(parent_key, rule_node_key, {"cost": rule.cost})

        has_constraints = (
            rule.prev_classes
            or rule.prev_tokens
            or rule.next_tokens
            or rule.next_classes
        )
        if has_constraints:
            constraints = {}
            if rule.prev_classes:
                constraints["prev_classes"] = rule.prev_classes
            if rule.prev_tokens:
                constraints["prev_tokens"] = rule.prev_tokens
            if rule.next_tokens:
                constraints["next_tokens"] = rule.next_tokens
            if rule.next_classes:
                constraints["next_classes"] = rule.next_classes
            edge_to_rule["constraints"] = constraints

    for node_key, node in enumerate(graph.node):
        ordered_children = {}
        rule_children_keys = node.get("rule_children")

        # add rule children to ordered_children dict under '__rules__''
        if rule_children_keys:
            ordered_children["__rules__"] = sorted(
                rule_children_keys, key=lambda x: graph.edge[node_key][x]["cost"]
            )
            node.pop("rule_children")

        token_children = node.get("token_children")

        # add token children to ordered_children dict by token
        if token_children:
            for token, token_key in token_children.items():
                ordered_children[token] = [token_key]

                # add rule children there as well
                if rule_children_keys:
                    ordered_children[token] += rule_children_keys

                # sort both by weight
                ordered_children[token].sort(
                    key=lambda _new_tkn_key: graph.edge[node_key][_new_tkn_key]["cost"]
                )
            # remove token children
            node.pop("token_children")
        node["ordered_children"] = ordered_children

    return graph
